/*
 *  This file is part of libosc++.
 *
 *  libosc++ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libosc++ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesse General Public License
 *  along with libosc++.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright 2009 Cristian Maglie
 *
 *  Cristian Maglie <liboscpp@bug.st>
 *  http://bug.st/libosc++
 */

#ifndef BUNDLE_H_INDCLUDED
#define BUNDLE_H_INDCLUDED

#include <libosc++/exception.h>
#include <libosc++/message.h>
#include <libosc++/bundle.h>

#include <string>
#include <boost/shared_ptr.hpp>

using namespace std;
using namespace boost;

namespace osc
{

/**
 * An OSC bundle class.
 * This class holds OSC bundles and can be used to construct
 * a new one or decode an external buffer.
 */
class Bundle
{
public:
	Bundle();
	~Bundle();

	/**
	 * Clear the contents of the bundle.
	 * After a call to this method the old bundle content is discarded and
	 * a new bundle can be composed or decoded.
	 */
	void clear();

	/**
	 * Starts a new bundle.
	 * Starts a new OSC bundle that must be closed with end() method.
	 * @param _time The OSC timestampe of the message. If not specified the <b>Immediate</b> timestamp (1) is used.
	 * @throws BundleAlreadyOpenedException If another bundle was already
	 *   started or decoded.
	 * @see end()
	 */
	void begin(const uint64_t _time = 1);

	/**
	 * Append a message to the actual bundle.
	 * The actual bundle must be already started with <b>begin()</b>.
	 * @param _v The message to append.
	 * @throws BundleNotOpenedExecption If actual bundle is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeMessage(Message &_v);

	/**
	 * Append a bundle to the actual bundle.
	 * The actual bundle must be already started with <b>begin()</b>.
	 * @param _v The bundle to append.
	 * @throws BundleNotOpenedExecption If actual bundle is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeBundle(Bundle &_v);

	/**
	 * Append raw data to the actual bundle.
	 * The actual bundle must be already started with <b>begin()</b>.
	 * This function must be used with careful because it doesn't
	 * validate inserted data.
	 * @param _v A std::string of data to append.
	 * @throws BundleNotOpenedExecption If actual bundle is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeRaw(const string &_v);

	/**
	 * Close the composing of the current bundle.
	 * After a call on this method, the bundle is complete and can be
	 * queried or transferred in binary form.
	 * @throws BundleNotOpenedExecption If actual bundle is not started
	 *   with <b>begin()</b> method.
	 */
	void end();

	/**
	 * Query the size of the bundle.
	 * @returns The size in bytes of the actual OSC bundle.
	 * @throws BundleNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method.
	 * @see end()
	 */
	int32_t size();

	/**
	 * Copyies the bundle into an external buffer.
	 * @param _buffer The buffer where the bundle will be copied. The buffer
	 *   must be preallocated with enough capacity to hold the entire bundle.
	 *   The bundle size can be obtained in advance using <b>size()</b> method.
	 * @throws BundleNotClosedException If the actual bundle was not
	 *   closed with <b>end()</b> method.
	 * @see size(), end()
	 */
	void output(char *_buffer);

	/**
	 * Copyies the bundle into an external buffer.
	 * @param _buffer A string where the bundle will be appended.
	 * @throws BundleNotClosedException If the actual bundle was not
	 *   closed with <b>end()</b> method.
	 * @see end()
	 */
	void output(string &_buffer);

private:
	bool started;
	bool closed;

	string data;
};

/**
 * Shared pointer for Bundle class
 */
typedef shared_ptr<Bundle> BundlePtr;

/**
 * "Bundle Already Opened" Exception class
 */
class BundleAlreadyOpenedException: public Exception
{
public:
	BundleAlreadyOpenedException() :
		Exception("Bundle already opened.") { }
};

/**
 * "Bundle Not Opened" Execption class
 */
class BundleNotOpenedExecption: public Exception
{
public:
	BundleNotOpenedExecption() :
		Exception("Bundle not opened.") { }
};

/**
 * "Bundle Not Closed" Exception class
 */
class BundleNotClosedException: public Exception
{
public:
	BundleNotClosedException() :
		Exception("Bundle not closed.") { }
};

} // namespace osc

#endif // BUNDLE_H_INDCLUDED
