/*
 *  This file is part of libosc++.
 *
 *  libosc++ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libosc++ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesse General Public License
 *  along with libosc++.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright 2009 Cristian Maglie
 *
 *  Cristian Maglie <liboscpp@bug.st>
 *  http://bug.st/libosc++
 */

#ifndef MESSAGE_H_INDCLUDED
#define MESSAGE_H_INDCLUDED

#include <libosc++/exception.h>
#include <libosc++/osctypes.h>

#include <boost/shared_ptr.hpp>

#include <string>

using namespace std;
using namespace boost;

namespace osc {

/**
 * An OSC message class.
 * This class holds all the content of an OSC message.
 *
 * Can be used either to decode incoming OSC binary data or to
 * compose new OSC message.
 */
class Message {
public:
	Message();
	~Message();

	/**
	 * Clear the contents of the message.
	 * After a call to this method the old message content is discarded and
	 * a new message can be composed or decoded.
	 */
	void clear();

	/**
	 * Starts a new message.
	 * Starts a new OSC message that must be closed with end() method.
	 * @param _address The OSC address of the message. If address doesn't begin
	 *   with <b>'/'</b> an additional slash is added at the beginning.
	 * @throws MessageAlreadyOpenedExecption If another message was already
	 *   started or decoded.
	 * @see end()
	 */
	void begin(const string &_address);

	/**
	 * Append a 32 bit integer parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeInt32(const int32_t _v);

	/**
	 * Append a 64 bit integer parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeInt64(const int64_t _v);

	/**
	 * Append a 32 bit floating point parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeFloat32(const float _v);

	/**
	 * Append a 64 bit floating point parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeFloat64(const double _v);

	/**
	 * Append a string parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeString(const char *_v);

	/**
	 * Append a string parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * The string passed must NOT contain '\0' character.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeString(const string &_v) { writeString(_v.c_str()); }

	/**
	 * Append a blob parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The buffer with the contents to insert in the message.
	 * @param _size The size of blob buffer in bytes.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeBlob(const char *_v, const int32_t _size);

	/**
	 * Append a boolean parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeBool(const bool _v);

	/**
	 * Append a Nil value parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeNil();

	/**
	 * Append an Infinitum value parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeInf();

	/**
	 * Append a char parameter to the message.
	 * The message must be already started with <b>begin()</b>.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see begin()
	 */
	void writeChar(const char _v);

	/**
	 * Append a 32 bit integer parameter to the message.
	 * It's the overloaded version of <b>writeInt32()</b> method.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see writeInt32()
	 */
	void write(const int32_t &_v) { writeInt32(_v); }

	/**
	 * Append a 32 bit floating point parameter to the message.
	 * It's the overloaded version of <b>writeFloat32()</b> method.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see writeFloat32()
	 */
	void write(const float &_v) { writeFloat32(_v); }

	/**
	 * Append a 64 bit floating point parameter to the message.
	 * It's the overloaded version of <b>writeFloat64()</b> method.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see writeFloat64()
	 */
	void write(const double &_v) { writeFloat64(_v); }

	/**
	 * Append a string parameter to the message.
	 * It's the overloaded version of <b>writeString()</b> method.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see writeString()
	 */
	void write(const char *&_v) { writeString(_v); }

	/**
	 * Append a string parameter to the message.
	 * It's the overloaded version of <b>writeString()</b> method.
	 * @param _v The value to insert in the message.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 * @see writeString()
	 */
	void write(const string &_v) { writeString(_v); }

	/**
	 * Close the composing of the current message.
	 * After a call on this method, the message is complete and can be
	 * queried or transferred in binary form.
	 * @throws MessageNotOpenedExecption If actual message is not started
	 *   with <b>begin()</b> method.
	 */
	void end();

	/**
	 * Query the status of the message.
	 * @returns <b>true</b> if the message is closed and can be queried
	 *   or transferred; <b>false</b> if the message is empty, or it's
	 *   in composing phase.
	 */
	bool isComplete() const { return closed; };

	/**
	 * Query the size of the message.
	 * @returns The size in bytes of the actual OSC message.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method.
	 * @see end()
	 */
	int size() const;

	/**
	 * Copyies the message into an external buffer.
	 * @param _buffer The buffer where the messagge will be copied. The buffer
	 *   must be preallocated with enough capacity to hold the entire message.
	 *   The message size can be obtained in advance using <b>size()</b> method.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method.
	 * @see size(), end()
	 */
	void output(char *_buffer) const;

	/**
	 * Copyies the message into an external buffer.
	 * @param _buffer A string where the messagge will be appended.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method.
	 * @see end()
	 */
	void output(string &_buffer) const;

	/**
	 * Decodes an OSC message in an external buffer.
	 * The actual message is discarded and substituted with the message contained
	 * in the external buffer.
	 * @param _buffer The buffer with the message to decode.
	 * @param _len The size of the buffer in bytes.
	 * @throws InvalidMessagePacketException If <b>_buffer</b> contains an
	 *   invalid or corrupted OSC message.
	 */
	void input(const char *_buffer, const int _len);

	/**
	 * Decodes an OSC message in an external buffer.
	 * The actual message is discarded and substituted with the message contained
	 * in the external buffer.
	 * @param _buffer A string containing the message to decode.
	 * @throws InvalidMessagePacketException If <b>_buffer</b> contains an
	 *   invalid or corrupted OSC message.
	 */
	void input(const string &_buffer);

	/**
	 * Query the OSC address of actual message.
	 * @returns A string with the OSC address of actual message.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @see end(), input()
	 */
	string getAddress() const;

	/**
	 * Query the OSC typetag sequence of actual message.
	 * @returns A string with the OSC typetag sequence of actual message without
	 *   the <b>','</b> charater at beginning.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @see end(), input()
	 */
	string getTag() const;

	/**
	 * Set the arguments pointer to the first parameter of the actual message.
	 * After the call to this method any subsequent call to any of the <b>get..()</b>
	 * methods will return the first parameter contained in the actual message.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @see end(), input()
	 */
	void restart();

	/**
	 * Reads a 32 bits integer from actual argument pointer.
	 * The argument pointer is incremented after read.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 * @return The value read from message at argument pointer.
	 */
	int32_t getInt32();

	/**
	 * Reads a 64 bits integer from actual argument pointer.
	 * The argument pointer is incremented after read.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 * @return The value read from message at argument pointer.
	 */
	int64_t getInt64();

	/**
	 * Reads a 32 bits floating point from actual argument pointer.
	 * The argument pointer is incremented after read.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 * @return The value read from message at argument pointer.
	 */
	float getFloat32();

	/**
	 * Reads a 64 bits floating point from actual argument pointer.
	 * The argument pointer is incremented after read.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 * @return The value read from message at argument pointer.
	 */
	double getFloat64();

	/**
	 * Reads a string from actual argument pointer.
	 * The argument pointer is incremented after read.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 * @return The value read from message at argument pointer.
	 */
	string getString();

	/**
	 * Reads a 32 bits integer from actual argument pointer.
	 * It's the overloaded version of <b>getInt32()</b> method.
	 * @param _r The variabile where the parameter value is put in.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 */
	void get(int32_t &_r) { _r = getInt32(); }

	/**
	 * Reads a 64 bits integer from actual argument pointer.
	 * It's the overloaded version of <b>getInt64()</b> method.
	 * @param _r The variabile where the parameter value is put in.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 */
	void get(int64_t &_r) { _r = getInt64(); }

	/**
	 * Reads a 32 bits floating point from actual argument pointer.
	 * It's the overloaded version of <b>getFloat32()</b> method.
	 * @param _r The variabile where the parameter value is put in.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 */
	void get(float &_r) { _r = getFloat32(); }

	/**
	 * Reads a 64 bits floating point from actual argument pointer.
	 * It's the overloaded version of <b>getFloat64()</b> method.
	 * @param _r The variabile where the parameter value is put in.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 */
	void get(double &_r) { _r = getFloat64(); }

	/**
	 * Reads a string from actual argument pointer.
	 * It's the overloaded version of <b>getString()</b> method.
	 * @param _r The variabile where the parameter value is put in.
	 * @throws MessageNotClosedException If the actual message was not
	 *   closed with <b>end()</b> method, or the last <b>input()</b> call
	 *   was unsuccessfull.
	 * @throws EndOfMessageReachedException If the arguments pointer is at
	 *   the end of message.
	 * @throws TypesDoesNotCorrespondsInReadException If the argument is not
	 *   of the expected type
	 */
	void get(string &_r) { _r = getString(); }

	/**
	 * Check if the argument pointer reached the end of message.
	 * @returns <b>true</b> if the argument pointer is at the end
	 *   of parameters list, <b>false</b> otherwise.
	 */
	bool endOfMessage() const { return eom; }

private:
	bool started;
	bool closed;

	string header;
	string tag;
	string args;

	bool eom;
	int tagPointer;
	void checkTag(const char _tag);
	int argPointer;
};

/**
 * Shared pointer for Message class
 */
typedef shared_ptr<Message> MessagePtr;

/**
 * "Invalid Address" Exception class
 */
class InvalidAddressException : public Exception {
public:
	InvalidAddressException() :
		Exception("Invalid address.") { }
};

/**
 * "Message Not Opened" Execption class
 */
class MessageNotOpenedExecption : public Exception {
public:
	MessageNotOpenedExecption() :
		Exception("Message not opened.") { }
};

/**
 * "Message Already Opened" Execption class
 */
class MessageAlreadyOpenedExecption : public Exception {
public:
	MessageAlreadyOpenedExecption() :
		Exception("Message is already opened.") { }
};

/**
 * "Message Not Closed" Exception class
 */
class MessageNotClosedException : public Exception {
public:
	MessageNotClosedException() :
		Exception("Message not closed.") { }
};

/**
 * "Invalid Message Packet" Exception class
 */
class InvalidMessagePacketException : public Exception {
public:
	InvalidMessagePacketException() :
		Exception("Invalid message packet.") { }
	/**
	 * Construct the exception with a given message.
	 * @param _msg The exception message.
	 */
	InvalidMessagePacketException(const char *_msg) : Exception(_msg) { }
};

/**
 * "End Of Message Reached" Exception class
 */
class EndOfMessageReachedException : public Exception {
public:
	EndOfMessageReachedException() :
		Exception("End of message reached.") { }
};

/**
 * "Types Does Not Corresponds In Read" Exception class
 */
class TypesDoesNotCorrespondsInReadException : public Exception {
public:
	TypesDoesNotCorrespondsInReadException() :
		Exception("Type of requested parameter differs from the one contained in OSC packet.") { }
};

} // namespace osc

#endif // MESSAGE_H_INDCLUDED
