/*
 *  This file is part of libosc++.
 *
 *  libosc++ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libosc++ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesse General Public License
 *  along with libosc++.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright 2009 Cristian Maglie
 *
 *  Cristian Maglie <liboscpp@bug.st>
 *  http://bug.st/libosc++
 */

#include <libosc++/message.h>
#include <endian_utils.h>

namespace osc {

Message::Message() :
	started(false),
	closed(false),
	header(""),
	tag(","),
	args("")
{
	// Empty
}

Message::~Message() 
{
	// Empty
}

void Message::clear()
{
	header = "";
	tag = ",";
	args = "";
	started = false;
	closed = false;
}

void Message::begin(const string &_address)
{
	if (started || closed)
		throw MessageAlreadyOpenedExecption();
	if (_address.size()<1)
		throw InvalidAddressException();
	if (_address[0] != '/')
		header = '/';
	header += _address;
	header += '\0';
	align32(header);
	started = true;
}

void Message::writeInt32(const int32_t _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	appendInt32(args, _v);
	tag += INT32_TAG;
}

void Message::writeInt64(const int64_t _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	appendInt64(args, _v);
	tag += INT64_TAG;
}

void Message::writeFloat32(const float _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	appendFloat32(args, _v);
	tag += FLOAT32_TAG;
}

void Message::writeFloat64(const double _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	appendFloat64(args, _v);
	tag += FLOAT64_TAG;
}

void Message::writeString(const char *_v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	args += _v;
	args += '\0';
	align32(args);
	tag += STRING_TAG;
}

void Message::writeBlob(const char *_v, const int32_t _size)
{
	if (!started)
		throw MessageNotOpenedExecption();
	appendInt32(args, _size);
	args.append(_v, _size);
	align32(args);
	tag += BLOB_TAG;
}

void Message::writeBool(const bool _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	tag += _v ? TRUE_TAG : FALSE_TAG;
}

void Message::writeNil()
{
	if (!started)
		throw MessageNotOpenedExecption();
	tag += NIL_TAG;
}

void Message::writeInf()
{
	if (!started)
		throw MessageNotOpenedExecption();
	tag += INF_TAG;
}

void Message::writeChar(const char _v)
{
	if (!started)
		throw MessageNotOpenedExecption();
	args += '\0';
	args += '\0';
	args += '\0';
	args += _v;
	tag += CHAR_TAG;
}

void Message::end()
{
	if (!started)
		throw MessageNotOpenedExecption();
	tag += '\0';
	align32(tag);
	started = false;
	closed = true;
	restart();
}

int Message::size() const
{
	if (!closed)
		throw MessageNotClosedException();
	return header.size() + tag.size() + args.size();
}

void Message::output(char *_buffer) const
{
	if (!closed)
		throw MessageNotClosedException();

	header.copy(_buffer, header.size());
	_buffer += header.size();
	tag.copy(_buffer, tag.size());
	_buffer += tag.size();
	args.copy(_buffer, args.size());
}

void Message::output(string &_buffer) const
{
	if (!closed)
		throw MessageNotClosedException();
	_buffer.append(header);
	_buffer.append(tag);
	_buffer.append(args);
}

static inline char next(const char *&_buff, int &_l) {
	if (!_l)
		throw InvalidMessagePacketException("Premature end of packet");
	_l--;
	return *(_buff++);
}

static string readOSCString(const char *&_buff, int &_l)
{
	string r = "";
	char c;
	do {
		c = next(_buff, _l);
		r += c;
	} while (c);
	while (r.size() % 4) {
		if (*_buff)
			throw InvalidMessagePacketException("Invalid string alignment found");
		r += next(_buff, _l);
	}
	return r;
}

void Message::input(const char *_buffer, const int _len)
{
	int l = _len;

	clear();

	// Retrieve address
	header = readOSCString(_buffer, l);
	if (header[0] != '/')
		throw InvalidMessagePacketException("Packet address must start with '/'");

	// Retrieve type tags
	tag = readOSCString(_buffer, l);
	if (tag[0] != ',')
		throw InvalidMessagePacketException("Packet type-tag must start with ','");

	// Retrieve payload
	args.append(_buffer, l);
	if (args.size() % 4 != 0)
		throw InvalidMessagePacketException("Invalid argument alignment found");

	closed = true;
	restart();
}

string trim(string _v)
{
	// FIXME: Can be done in a much better way....
	int l = _v.length();
	while (l && _v.at(0) == '\0') {
		_v = _v.substr(1);
		l--;
	}
	while (l && _v.at(l-1) == '\0') {
		_v = _v.substr(0, l-1);
		l--;
	}
	return _v;
}

string Message::getAddress() const
{
	if (!closed)
		throw MessageNotClosedException();
	return trim(header);
}

string Message::getTag() const
{
	if (!closed)
		throw MessageNotClosedException();
	return trim(tag.substr(1));
}

static string readOSCStringFromSTLString(string &_buff, int &_pos)
{
	string r = "";
	char c = _buff.at(_pos++);
	while (c) {
		r += c;
		c = _buff.at(_pos++);
	};
	int rem = ((r.size() + 1) % 4);
	if (rem)
		_pos += 4 - rem;
	return r;
}

void Message::restart()
{
	if (!closed)
		throw MessageNotClosedException();
	tagPointer = 1;
	argPointer = 0;
	eom = false;
}

void Message::checkTag(const char _tag)
{
	if (!closed)
		throw MessageNotClosedException();
	if (eom)
		throw EndOfMessageReachedException();
	if (tag[tagPointer] != _tag)
		throw TypesDoesNotCorrespondsInReadException();
	tagPointer++;
	if (tag[tagPointer] == '\0')
		eom = true;
}

int32_t Message::getInt32()
{
	checkTag(INT32_TAG);
	return readInt32(args, argPointer);
}

int64_t Message::getInt64()
{
	checkTag(INT64_TAG);
	return readInt64(args, argPointer);
}

float Message::getFloat32()
{
	checkTag(FLOAT32_TAG);
	return readFloat32(args, argPointer);
}

double Message::getFloat64()
{
	checkTag(FLOAT64_TAG);
	return readFloat64(args, argPointer);
}

string Message::getString()
{
	checkTag(STRING_TAG);
	return readOSCStringFromSTLString(args, argPointer);
}

} // namespace osc
