/*
 *  This file is part of libosc++.
 *
 *  libosc++ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libosc++ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesse General Public License
 *  along with libosc++.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright 2009 Cristian Maglie
 *
 *  Cristian Maglie <liboscpp@bug.st>
 *  http://bug.st/libosc++
 */

#include <libosc++/oscostream.h>

#include <iostream>

using namespace std;

namespace osc {

oscostream::oscostream() : 
	messageOpened(false),
	bundleOpened(false),
	childrenBundleOpened(false),
	data("")
{
	// Empty
}

oscostream::~oscostream()
{
	// Empty
}

void oscostream::clear()
{
	bundleOpened = false;
	messageOpened = false;
	childrenBundleOpened = false;
	data = "";
}

unsigned int oscostream::size() const
{
	return data.size();
}

oscostream& oscostream::operator<<(const beginbundle &_v)
{
	if (messageOpened)
		throw MessageOpenedException();

	if (!bundleOpened) {
		bundle.reset(new Bundle());
		bundle->begin(_v.time);
		bundleOpened = true;
	} else if (!childrenBundleOpened) {
		children.reset(new oscostream());
		(*children) << _v;
		childrenBundleOpened = true;
	} else {
		(*children) << _v;
	}

	return *this;
}

oscostream& oscostream::operator<<(const beginmessage &_v)
{
	if (messageOpened)
		throw MessageOpenedException();

	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		message.reset(new Message());
		message->begin(_v.addr);
		messageOpened = true;
	}

	return *this;
}

oscostream& oscostream::operator<<(const close &_v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
		if (children->ready()) {
			bundle->writeRaw(children->data);
			children.reset();
			childrenBundleOpened = false;
		}

	} else if (messageOpened) {
		message->end();
		if (bundleOpened) {
			bundle->writeMessage(*message);
		} else {
			message->output(data);
		}
		message.reset();
		messageOpened = false;

	} else if (bundleOpened) {
		bundle->end();
		bundle->output(data);
		bundle.reset();
		bundleOpened = false;

	} else {
		throw BundleNotOpenedException();
	}

	return *this;
}

oscostream& oscostream::operator<<(const int32_t _v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeInt32(_v);
	}
	return *this;
}

oscostream& oscostream::operator<<(const bool _v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeBool(_v);
	}
	return *this;
}

oscostream& oscostream::operator<<(const nil &_v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeNil();
	}
	return *this;
}

oscostream& oscostream::operator<<(const inf &_v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeInf();
	}
	return *this;
}

oscostream& oscostream::operator<<(const char *_v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeString(_v);
	}
	return *this;
}

oscostream& oscostream::operator<<(const float _v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeFloat32(_v);
	}
	return *this;
}

oscostream& oscostream::operator<<(const double _v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeFloat64(_v);
	}
	return *this;
}

oscostream& oscostream::operator<<(const char _v)
{
	if (childrenBundleOpened) {
		(*children) << _v;
	} else {
		if (!messageOpened)
			throw MessageNotOpenedException();
		message->writeChar(_v);
	}
	return *this;
}

bool oscostream::ready() const
{
	return !messageOpened && !bundleOpened;
}

void oscostream::output(char *_buffer) const {
	if (messageOpened)
		throw MessageOpenedException();
	if (bundleOpened)
		throw BundleOpenedException();
	data.copy(_buffer, data.size());
}

void oscostream::output(string &_buffer) const {
	if (messageOpened)
		throw MessageOpenedException();
	if (bundleOpened)
		throw BundleOpenedException();
	_buffer.append(data);
}

} // namespace osc


