/*
 *  This file is part of libosc++.
 *
 *  libosc++ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libosc++ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesse General Public License
 *  along with libosc++.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Copyright 2009 Cristian Maglie
 *
 *  Cristian Maglie <liboscpp@bug.st>
 *  http://bug.st/libosc++
 */

#include <libosc++/udp.h>

#include <iostream>
#include <boost/array.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>

using namespace boost::asio;
using namespace std;

namespace osc {

UDPTransport::UDPTransport() : socket(io)
{
	socket.open(udp::v4());
}

UDPTransport::~UDPTransport()
{
	socket.close();
}

void UDPTransport::setRecvUDPEndpoint(string _addr, string _port)
{
	recvEndpoint = resolve(_addr, _port);
}

void UDPTransport::setRecvCB(function<void (MessagePtr)> _cb)
{
	recvCB = _cb;
}

udp::endpoint UDPTransport::resolve(string _addr, string _port)
{
	udp::resolver resolver(io);
	udp::resolver::query query(udp::v4(), _addr, _port);
	udp::endpoint receiver_endpoint = *resolver.resolve(query);
	return receiver_endpoint;
}

void UDPTransport::startRecvThread()
{
	thread(bind(&UDPTransport::run, this));
}

void UDPTransport::setSendUDPEndpoint(string _addr, string _port)
{
	sendEndpoint = resolve(_addr, _port);
}

void UDPTransport::run()
{
	try
	{
		socket.bind(recvEndpoint);

		while (true)
		{
			array<char, 2048> buff;
			udp::endpoint sender;
			size_t len = socket.receive_from(buffer(buff), sender);

			try
			{
				MessagePtr msg(new Message());
				msg->input(buff.data(), len);
				recvCB(msg);

			} catch (InvalidMessagePacketException &e) {
				std::cerr << "Invalid OSC Packet: " << e.what() << std::endl;
			}
		}

	} catch (std::exception& e) {
		std::cerr << "UDP Error: " << e.what() << std::endl;
	}
}

void UDPTransport::send(MessagePtr _m)
{
	try
	{
		string buff;
		_m->output(buff);
		socket.send_to(buffer(buff), sendEndpoint);

	} catch (std::exception& e) {
		std::cerr << "UDP Send Error: " << e.what() << std::endl;
	}
}

} // namespace osc
